#include "rt/RtAudio.h"
#include "rt/RtMidi.h"
#include <stdlib.h>
#include <math.h>
#include <iostream>
#include <string.h>
#include <cstdlib>
#include <vector>

#include <cstring>
#include "osc/OscOutboundPacketStream.h"
#include "ip/UdpSocket.h"
#include "ip/IpEndpointName.h"
#include "osc/OscReceivedElements.h"
#include "osc/OscPacketListener.h"

#define IP_MTU_SIZE 1536

#define SAMPLE double
#define MY_PIE 3.14159265358979
#define MY_SRATE 48000

unsigned int maxDelayLength = 100000;
float testFreak = 200;
int activeNotes = 0;
int initialized = 0;
int addNoise = 0;
int numberOfVoices = 150; // polyphony

// Set up OSC addresses, ports
char *hostName = NULL;
int outPort;
int inPort;

// Our structure to store individual notes
struct strangThang {
	double * delayBuffer;
	double strangFreak; // delay frequency
	double attenFaktor;
	
	int writeHead;
	int readHead;
	int pluck; // if pluck = 1, add noise to the buffer
	int inUse;
};

// Where to store all the currently-playing notes
std::vector <strangThang> noteVector (numberOfVoices); // how much polyphony


// simple frequency to numSamples function
unsigned int freakToSamps(float hz)
{
	int samps = (int) MY_SRATE / hz;
	return samps;
}

// simple midi to frequency function
double mtof(int note)
{
	double frequency = 440 * pow(2.0,(note-69.0)/12.0);
	return frequency;
}


void playNote( int midiPitch, int midiVelocity )
{
	double newFreak = mtof(midiPitch);
	int newNotePosition = 0; // newNotePosition is the address of the noteVector of our new note
	int more = 1;
	while (noteVector[newNotePosition].inUse) {
		newNotePosition++;
		if (newNotePosition == numberOfVoices) {
			more = 0;
		}
	}
	if (more) {
		noteVector[newNotePosition].strangFreak = newFreak;
		noteVector[newNotePosition].readHead = 0;
		noteVector[newNotePosition].writeHead =  freakToSamps(newFreak);
		noteVector[newNotePosition].attenFaktor = midiVelocity / 127.1;
		noteVector[newNotePosition].inUse = 1;
		noteVector[newNotePosition].pluck = 1;
		activeNotes++;
	}
	
}


////////////////////////////////////////////////////////////////
namespace osc{
	
	class OscReceiveTestPacketListener : public OscPacketListener{
	protected:
		
		void ProcessMessage( const osc::ReceivedMessage& m, const IpEndpointName& remoteEndpoint )
		{
			
			try {
				ReceivedMessageArgumentStream args = m.ArgumentStream();
				ReceivedMessage::const_iterator arg = m.ArgumentsBegin();
				
				if( strcmp( m.AddressPattern(), "/midi" ) == 0 ){
					
					osc::int32 pitch;
					osc::int32 velo;
					args >> pitch >> velo >> osc::EndMessage;
					
					std::cout << "received pitch "
					<< pitch << ", velocity " << velo << "\n";
					playNote( pitch, velo );
					
				}else{
					std::cout << "unrecognised address pattern: "
					<< m.AddressPattern() << "\n";
				}
				
			}catch( Exception& e ){
				std::cout << "error while parsing message: "
				<< m.AddressPattern() << ": " << e.what() << "\n";
			}
		}    
	};
	
	
	void SendMyNote( int midiPitch, int midiVelocity )
	{
		IpEndpointName host( hostName, outPort );
		
		char hostIpAddress[ IpEndpointName::ADDRESS_STRING_LENGTH ];
		host.AddressAsString( hostIpAddress );
		
		char buffer[IP_MTU_SIZE];
		osc::OutboundPacketStream p( buffer, IP_MTU_SIZE );
		UdpTransmitSocket socket( host );
		
		p.Clear();
		p << osc::BeginMessage( "/midi" )
		<< midiPitch << midiVelocity << osc::EndMessage;
		socket.Send( p.Data(), p.Size() );
	}
	
} // namespace osc
////////////////////////////////////////////////////////////////



void usage( void ) {
	std::cerr << "\nuseage:	./NetPlucker <ipaddress> <inputPortNum> <outputPortNum>\n";
	std::cerr << "	ipaddress: IP Address of the computer you're connecting to.\n";
	std::cerr << "	inputPortNum: Listen port number (must match other computer's output port number).\n";
	std::cerr << "	outputPortNum: Sending port number (must match other computer's input port number).\n\n";	
	exit( 1 );
}

// The midi callback function
void getNotez( double deltatime, std::vector< unsigned char > *message, void *userData )
{	
	unsigned int nBytes = message->size();
	
	if (nBytes == 3) {
		if ((int)message->at(2) != 0 && (int)message->at(0) == 144) { // only look at note on messages with a non-zero velocity
			int midiPitch = (int)message->at(1);
			int midiVelocity = (int)message->at(2);
			playNote( midiPitch, midiVelocity);
			osc::SendMyNote( midiPitch, midiVelocity );
		}
	}
}




// The audio callback function
int makeNoize( void *outputBuffer, void *inputBuffer, unsigned int nBufferFrames,
		double streamTime, RtAudioStreamStatus status, void *userData )
{	
//	double * inBuff = (double *) inputBuffer;
	double * outBuff = (double *) outputBuffer;

	for ( unsigned int j=0 ; j<nBufferFrames ; j++ ) { // first clear the output buffer
		outBuff[j] = 0;
	}
	
	std::vector<strangThang> * notes = (std::vector<strangThang> *) userData;

	
	for ( unsigned int i = 0 ; i < (*notes).size(); i++ ) { // for all notes stored in noteArray
		if ((*notes)[i].inUse) {
			double ampAverage = 0;
			
			// Noise Generator
			if ((*notes)[i].pluck) {
				for ( unsigned int k = 0; k < maxDelayLength + 1; k++ ) { // for all samples in the delay buffer
					
					(*notes)[i].delayBuffer[k] = ((rand() / (double) RAND_MAX) * 2)  - 1;
				}
				(*notes)[i].pluck = 0;
			}			
			
			
			for ( unsigned int j=0 ; j<nBufferFrames ; j++ ) { // for all samples stored in the delay buffer
				(*notes)[i].delayBuffer[(*notes)[i].writeHead] = (.5 * ((*notes)[i].delayBuffer[(*notes)[i].readHead] + ((*notes)[i].delayBuffer[((*notes)[i].readHead - 1) % maxDelayLength])) * (*notes)[i].attenFaktor);
				(*notes)[i].writeHead = ((*notes)[i].writeHead + 1) % (maxDelayLength + 1); // advance writehead

				ampAverage += fabs((*notes)[i].delayBuffer[(*notes)[i].readHead]);
				outBuff[j] += (*notes)[i].delayBuffer[(*notes)[i].readHead]; // store inst 1's delay buffer into audio output
				(*notes)[i].readHead = ((*notes)[i].readHead + 1) % (maxDelayLength + 1); // advance readhead
			}
			
			ampAverage = ampAverage / nBufferFrames;
			if (ampAverage < pow(10, -3)) {
				(*notes)[i].inUse = 0;
				activeNotes--;
			}
		}
	}

	return 0;
}



// main function
int main( int argc, char ** argv )
{
	// Command Line parsing
	//---------------------
	if( argc == 2 && strcmp(argv[1], "-help") == 0)
	{
		usage();
		exit(0);
	}
	else if ( argc == 4 ) {
		
		hostName = argv[1];
		inPort = atoi(argv[2]);
		outPort = atoi(argv[3]);
	}
	else {
		std::cout << "Here, let me help you out there:" << std::endl;
		usage();
		exit(0);
	}


	
	// Initialize Note Vector for strangThang Objects
	//----------------------------------------------- 	
	for (int i = 0; i <numberOfVoices; i++) {
		noteVector[i].delayBuffer = new double[maxDelayLength];
		for (unsigned int j = 0; j < maxDelayLength; j++) {
			noteVector[i].delayBuffer[j] = 0;
		}
	}
	
	noteVector[0].writeHead =  freakToSamps(testFreak);	
	
	// Set up RtAudio junk
	//--------------------
	RtAudio dac;
	RtAudio::DeviceInfo info;
	unsigned int devices = dac.getDeviceCount();
	unsigned int w = 0;
	info = dac.getDeviceInfo(w);
	while ( w < devices && !info.isDefaultOutput) {
		w++;
		info = dac.getDeviceInfo(w);
	}
		
	if ( devices < 1 ) {
		std::cout << "\nNo audio devices found!\n";
		exit( 0 );
	}
	
	RtAudio::StreamParameters Oparameters, Iparameters;
	Oparameters.deviceId = w;
	Oparameters.nChannels = 1;
	Oparameters.firstChannel = 0;
	Iparameters.deviceId = 0;
	Iparameters.nChannels = 1;
	unsigned int sampleRate = 48000;
	unsigned int bufferFrames = 256; // 256 sample frames

	
	RtAudio::StreamOptions options;
	
	try {
		dac.openStream( &Oparameters, &Iparameters, RTAUDIO_FLOAT64,
					   sampleRate, &bufferFrames, &makeNoize, (void *) &noteVector /*(void *) noteVector*/ );
		dac.startStream();
	}
	catch ( RtError& e ) {
		e.printMessage();
		exit( 0 );
	}
	
		
	// Set up OSC Reciever
    osc::OscReceiveTestPacketListener listener;
	UdpListeningReceiveSocket s(
								IpEndpointName( IpEndpointName::ANY_ADDRESS, inPort ),
								&listener );
	
		
	// Set up RtMidi Junk	
	//-------------------
	RtMidiIn *midiin = 0;
	// RtMidiIn constructor
	try {
		midiin = new RtMidiIn();
	}
	catch ( RtError &error ) {
		error.printMessage();
		exit( EXIT_FAILURE );
	}
	// Check available ports vs. specified.
	unsigned int nPorts = midiin->getPortCount();
	if ( nPorts > 0) {
		try {
			midiin->openPort( nPorts - 1 );
		}
		catch ( RtError &error ) {
			error.printMessage();
			goto cleanup;
		}
		midiin->setCallback( &getNotez , (void *) &inPort );
		std::cout << "\n\nMidi keyboard found, play away!" << std::endl;
		std::cout << "press ctrl-c to end program.\n" << std::endl;
	}
	else {
		std::cerr << "I can't seem to find any midi ports...\n" << std::endl;
		goto cleanup;
	}
	
	
	
	s.RunUntilSigInt();	


	// Clean up
cleanup:
	try {
		// Stop the stream
		dac.stopStream();
	}
	catch (RtError& e) {
		e.printMessage();
	}
	
	if ( dac.isStreamOpen() ) dac.closeStream();
	delete midiin;

	return 0;
}
